/* image.c -
 *
 * Copyright (C) 2001  Jochen Voss.  */

static const  char  rcsid[] = "$Id: image.c,v 1.2 2001/09/17 22:58:57 voss Rel $";

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>

#include <glib.h>

#include <SDL.h>
#include <SDL_image.h>

#include "graphcrypt.h"


static  SDL_Surface *gray;	/* full grayscale image */
static  SDL_Surface *tmpl;	/* approximation with reduced colors */


static void
copy_image_gray (SDL_Surface *image)
{
  SDL_Color colors[256];
  int  i, res;
 
  tmpl = SDL_CreateRGBSurface (SDL_SWSURFACE, image->w, image->h, 8,
			       0, 0, 0, 0);
  if (! tmpl) {
    fprintf (stderr, "cannot allocate reduced image: %s\n", SDL_GetError ());
    exit (1);
  }
  for (i=0; i<256; ++i) {
    colors[i].r = colors[i].g = colors[i].b = i;
  }
  res = SDL_SetColors (tmpl, colors, 0, 256);
  if (! res) {
    fprintf (stderr, "cannot set colormap: %s\n",
	     SDL_GetError ());
  }

  gray = SDL_ConvertSurface (image, tmpl->format, SDL_SWSURFACE);
  if (! gray) {
    fprintf (stderr, "cannot convert image to grayscale: %s\n",
	     SDL_GetError ());
    exit (1);
  }
}

void
handle_image (const char *infile, const char *outfile)
{
  SDL_Surface *screen, *image;
  struct image  src, dest;
  int  done;
  SDL_Event  event;

  /* Initialize the SDL library */
  if (SDL_Init (SDL_INIT_VIDEO) < 0) {
    g_error ("Couldn't initialize SDL: %s", SDL_GetError ());
  }

  /* Open the image file */
  printf ("loading image \"%s\" ...", infile);
  fflush (NULL);
  image = IMG_Load (infile);
  if (image == NULL) {
    puts (" FAILED");
    g_error ("cannot load %s", infile);
  } else {
    puts (" done");
  }
  if (image->w > page_width() || image->h > page_height()) {
    g_error ("image size %dx%d is too large (maximum is %dx%d)",
	     image->w, image->h, page_width(), page_height());
  }
  SDL_WM_SetCaption (infile, "showimage");

  printf ("converting image to grayscale ...");
  fflush (NULL);
  copy_image_gray (image);
  src.data = gray->pixels;
  src.w = gray->w;
  src.h = gray->h;
  src.pitch = gray->pitch;
  prepare (&src);
  dest.data = tmpl->pixels;
  dest.w = tmpl->w;
  dest.h = tmpl->h;
  dest.pitch = tmpl->pitch;
  shade (&dest, &src);
  puts (" done");

  screen = SDL_SetVideoMode (tmpl->w, tmpl->h, 8,
			     SDL_SWSURFACE|SDL_ANYFORMAT);
  if (screen == NULL) {
    g_error ("Couldn't set %dx%d video mode", tmpl->w, tmpl->h);
  }

  /* Set the palette, if one exists */
  if (tmpl->format->palette) {
    SDL_SetColors (screen, tmpl->format->palette->colors,
		   0, tmpl->format->palette->ncolors);
  }

  /* Display the image */
  SDL_BlitSurface (tmpl, NULL, screen, NULL);
  SDL_UpdateRect (screen, 0, 0, 0, 0);

  puts ("shading image, valid keys are");
  puts ("  a     - use a regular raster");
  puts ("  0     - use an irregular raster");
  puts ("  1-9   - use an irregular raster with noise");
  puts ("          small numbers: much noise, converges fast");
  puts ("          large numbers: less noise, converges slowly");
  printf ("  q     - write encrypted image to \"%s\" and quit\n", outfile);
  printf ("  w     - write encrypted image to \"%s\" and continue\n", outfile);
  puts ("  x,ESC - abort without saving");

  done = 0;
  while (!done) {
    if (SDL_PollEvent (&event)) {
      switch (event.type) {
      case SDL_KEYUP:
	switch (event.key.keysym.sym) {
	case SDLK_q:
	  printf ("writing encrypted image to file \"%s\" ...", outfile);
	  fflush (NULL);
	  postscript_print_encrypted_image (outfile, infile, &dest);
	  puts (" done");
	case SDLK_x:
	case SDLK_ESCAPE:
	  done = 1;
	  break;
	default:
	  break;
	}
	break;
      case SDL_KEYDOWN:
	switch (event.key.keysym.sym) {
	case SDLK_a:
	  diffuse (&dest, &src);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_1:
	  gibbs_diffuse_rnd (&dest, &src, 0.0005);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_2:
	  gibbs_diffuse_rnd (&dest, &src, 0.001);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_3:
	  gibbs_diffuse_rnd (&dest, &src, 0.0015);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_4:
	  gibbs_diffuse_rnd (&dest, &src, 0.0025);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_5:
	  gibbs_diffuse_rnd (&dest, &src, 0.005);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_6:
	  gibbs_diffuse_rnd (&dest, &src, 0.010);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_7:
	  gibbs_diffuse_rnd (&dest, &src, 0.015);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_8:
	  gibbs_diffuse_rnd (&dest, &src, 0.018);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_9:
	  gibbs_diffuse_rnd (&dest, &src, 0.020);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_0:
	  diffuse_rnd (&dest, &src);
	  SDL_BlitSurface (tmpl, NULL, screen, NULL);
	  SDL_UpdateRect (screen, 0, 0, 0, 0);
	  break;
	case SDLK_w:
	  printf ("writing encrypted image to file \"%s\" ...", outfile);
	  fflush (NULL);
	  postscript_print_encrypted_image (outfile, infile, &dest);
	  puts (" done");
	  break;
	case SDLK_i:
	  printf ("writing shaded image to file \"%s\" ...", outfile);
	  fflush (NULL);
	  postscript_print_image (outfile, infile, &dest, 0, 0);
	  puts (" done");
	  break;
	default:
	  break;
	}
	break;
      case SDL_QUIT:
	done = 1;
	break;
      default:
	break;
      }
    } else {
      SDL_Delay (10);
    }
  }
  SDL_FreeSurface (tmpl);

  /* We're done! */
  SDL_Quit ();
}
