/* postscript.c - write the image into a PostScript file
 *
 * Copyright (C) 2001  Jochen Voss.  */

static const  char  rcsid[] = "$Id: postscript.c,v 1.2 2001/09/17 22:58:57 voss Rel $";

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include <glib.h>

#include "graphcrypt.h"


/* #define DINA3 1 */

/* Size in PostScript points: */
#if defined(DINA3)
#define  PAGEWIDTH   830
#define  PAGEHEIGHT 1138
#else
#define  PAGEWIDTH   583
#define  PAGEHEIGHT  790
#endif


static  int  pages;

static void
write_header (FILE *f, const char *title)
{
  time_t  t;

  time (&t);

  fputs ("%!PS-Adobe-3.0\n", f);
  fprintf (f, "%%%%Creator: " PACKAGE " " VERSION "\n");
  fprintf (f, "%%%%CreationDate: %s", ctime (&t));
  fprintf (f, "%%%%Title: %s\n", title);
  fputs ("%%LanguageLevel: 2\n", f);
#if defined(DINA3)
  fputs ("%%DocumentMedia: a3 842 1190 0 () ()\n", f);
#else
  fputs ("%%DocumentMedia: a4 595 842 0 () ()\n", f);
#endif
  fputs ("%%Pages: (atend)\n", f);
  fputs ("%%EndComments\n", f);
  fputs ("%%BeginSetup\n", f);
#if defined(DINA3)
  fputs ("<< /PageSize [842 1190] >> setpagedevice\n", f);
#else
  fputs ("<< /PageSize [595 842] >> setpagedevice\n", f);
#endif
  fputs ("%%EndSetup\n", f);

  pages = 0;
}

static  int  a85n;
static  int  a85m;
static  uchar  a85[4];

static void
a85start (void)
{
  a85n = 0;
  a85m = 0;
}

static void
a85putchar (int c, FILE *f)
{
  a85[a85n++] = c;
  if (a85n == 4) {
    unsigned long  val;
    int  i;
    val = a85[0]*0x1000000ul+a85[1]*0x10000ul+a85[2]*0x100ul+a85[3];
    if (val == 0) {
      fputc ('z', f);
    } else {
      uchar  buffer [5];
      for (i=0; i<5; ++i) {
	buffer[4-i] = '!' + val%85ul;
	val /= 85ul;
      }
      for (i=0; i<5; ++i)  fputc (buffer[i], f);
    }
    a85n = 0;

    ++a85m;
    if (a85m >= 15) {
      fputc ('\n', f);
      a85m = 0;
    }
  }
}

static void
a85flush (FILE *f)
{
  if (a85n > 0) {
    uchar  buffer [5];
    unsigned long  val;
    int  i;
    
    for (i=a85n; i<5; ++i)  a85[i] = 0;
    val = a85[0]*0x1000000ul+a85[1]*0x10000ul+a85[2]*0x100ul+a85[3];
    for (i=0; i<5; ++i) {
      buffer[4-i] = '!' + val%85ul;
      val /= 85ul;
    }
    for (i=0; i<a85n+1; ++i)  fputc (buffer[i], f);
  }
  fputs ("~>\n", f);
}

static int
count_lines (struct image *im)
{
  int  len;

  /* this must match `write_image_data'! */
  len = im->w * im->h;
  return  (len+59)/60;
}

static void
write_image_data (const char *title, struct image *im,
		  int xoffset, int yoffset, FILE *f)
{
  int  i, j;

  ++pages;
  fprintf (f, "%%%%Page: %d\n", pages);
  fputs ("/Times-Roman findfont 8 scalefont setfont\n", f);
  fprintf (f, "26 26 moveto (%s) show\n", title);
  fputs ("26 36 moveto (GraphCrypt " VERSION
	 " - Copyright 2001 Jochen Voss "
	 " -  http://www.mathematik.uni-kl.de/~wwwstoch/voss/privat.html)"
	 " show\n", f);
  fputs ("gsave\n", f);
  fprintf (f, "6 46 translate %g %g scale\n", 72.0/DPI, 72.0/DPI);
  fprintf (f, "[ %d 0 0 %d %d %d ] concat\n",
	   im->w, -im->h, xoffset, 2*page_height()-yoffset);
  fprintf (f, "%d %d 8 [ %d 0 0 %d 0 0 ]\n",
	   im->w, im->h, im->w, im->h);
  fputs ("currentfile /ASCII85Decode filter\n", f);
  fprintf (f, "%%%%BeginData: %d ASCII Lines\n", 1+count_lines (im));
  fputs ("image\n", f);

  a85start ();
  for (j=0; j<im->h; ++j) {
    for (i=0; i<im->w; ++i) {
      a85putchar (im->data[i+im->pitch*j], f);
    }
  }
  a85flush (f);

  fputs ("%%EndData\n", f);
  fputs ("grestore\n", f);
  fputs ("showpage\n", f);
  fputs ("%%EndPage\n", f);
}

static void
write_tail (FILE *f)
{
  fputs ("%%Trailer\n", f);
  fprintf (f, "%%%%Pages: %d\n", pages);
  fputs ("%%EOF\n", f);
}

int
page_width (void)
/* Get the maximum image width in (image) pixels.  */
{
  return  0.5*PAGEWIDTH*(double)DPI/72;
}

int
page_height (void)
/* Get the maximum image height in (image) pixels.  */
{
  return  0.5*PAGEHEIGHT*(double)DPI/72;
}

void
postscript_print_image (const char *filename, const char *title,
			struct image *im, int xoffset, int yoffset)
{
  FILE *f;

  f = fopen (filename, "w");
  if (! f) {
    g_error ("cannot open output file \"%s\": %s",
	     filename, strerror (errno));
  }
  write_header (f, title);
  write_image_data (title, im, xoffset, yoffset, f);
  write_tail (f);
  fclose (f);
}
