/* shade.c - use dithering to reduce the number of colors
 *
 * Copyright (C) 2001  Jochen Voss.  */

static const  char  rcsid[] = "$Id: shade.c,v 1.2 2001/09/17 22:58:57 voss Rel $";

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <math.h>
#include <glib.h>
#include <jvrand.h>

#include "graphcrypt.h"


#define  RANGE  2


static uchar
get_pixel (struct image *im, int i, int j)
{
  return  im->data[i+im->pitch*j];
}

static void
set_pixel (struct image *im, int i, int j, uchar val)
{
  im->data[i+im->pitch*j] = val;
}

static uchar
reduce (uchar x)
{
  if (x < 32)  return  0;
  if (x < 64+32)  return  64;
  return  128;
}

static  double *dest_mean = NULL;
static  double *src_mean = NULL;
static  double *weight = NULL;

static double *
mean_values (struct image *im)
{
  int  w = im->w;
  int  h = im->h;
  double *buffer = g_new (double, w*h);
  int  i, j;

  for (j=0; j<h; ++j) {
    for (i=0; i<w; ++i) {
      buffer [i+w*j] = 0;
    }
  }
  for (j=0; j<h; ++j) {
    for (i=0; i<w; ++i) {
      uchar  val = get_pixel (im, i, j);
      int  k, l;

      for (k = CLAMP(i-RANGE, 0, w); k < CLAMP(i+RANGE+1, 0, w); ++k) {
	for (l = CLAMP(j-RANGE, 0, h); l < CLAMP(j+RANGE+1, 0, h); ++l) {
	  buffer [k+w*l] += val;
	}
      }
    }
  }
  for (j=0; j<h; ++j) {
    for (i=0; i<w; ++i) {
      buffer [i+w*j] *= weight [i+w*j];
    }
  }
  return  buffer;
}

static double *
weights (int w, int h)
{
  double *buffer = g_new (double, w*h);
  int  i, j;
  int  a, b;

  for (j=0; j<h; ++j) {
    if (j < RANGE) {
      b = j+RANGE+1;
    } else if (j > h-RANGE-1) {
      b = (w-1-j)+RANGE+1;
    } else {
      b = 2*RANGE+1;
    }
    for (i=0; i<w; ++i) {
      if (i < RANGE) {
	a = i+RANGE+1;
      } else if (i > w-RANGE-1) {
	a = (w-1-i)+RANGE+1;
      } else {
	a = 2*RANGE+1;
      }
      buffer [i+w*j] = 1.0/(a*b);
    }
  }
  return  buffer;
}

static void
modify_mean (int w, int h, int i, int j, double delta)
{
  int  k, l;
  
  for (k = CLAMP(i-RANGE, 0, w); k < CLAMP(i+RANGE+1, 0, w); ++k) {
    for (l = CLAMP(j-RANGE, 0, h); l < CLAMP(j+RANGE+1, 0, h); ++l) {
      dest_mean [k+w*l] += delta * weight[k+w*l];
    }
  }
}

static double
dist_sq (double x, double y)
{
  double  a = x - y;
  return  a*a;
}

double
delta (struct image *dest, int i, int j, uchar x)
{
  uchar  old = get_pixel (dest, i, j);
  int  w = dest->w;
  int  h = dest->h;
  int  k, l;
  double  val;
  
  if (x != old)  modify_mean (w, h, i, j, x-(double)old);
  val = 0;
  for (k = CLAMP(i-RANGE, 0, i); k < CLAMP(i+RANGE+1, i, w); ++k) {
    for (l = CLAMP(j-RANGE, 0, j); l < CLAMP(j+RANGE+1, j, h); ++l) {
      val += dist_sq (src_mean[k+w*l], dest_mean[k+w*l]);
    }
  }
  if (x != old)  modify_mean (w, h, i, j, old-(double)x);

  return  val;
}

static void
set_pixel_and_mean (struct image *dest, int i, int j, uchar x)
{
  uchar  old = get_pixel (dest, i, j);

  modify_mean (dest->w, dest->h, i, j, x-(double)old);
  set_pixel (dest, i, j, x);
}

void
prepare (struct image *im)
{
  int  w, h;
  int  i, j, k, cut;
  int  hist[256];

  w = im->w;
  h = im->h;
  
  for (i=0; i<256; ++i)  hist[i] = 0;
  for (i=0; i<w; ++i) {
    for (j=0; j<h; ++j) {
      uchar c = get_pixel (im, i, j);
      ++hist[c];
    }
  }

  k = 0.1*w*h+0.5;
  for (cut=255; cut>128; --cut) {
    k -= hist[cut];
    if (k<0)  break;
  }
  
  for (i=0; i<w; ++i) {
    for (j=0; j<h; ++j) {
      double  x;

      x = get_pixel (im, i, j);
      x = 128 * x / cut;
      set_pixel (im, i, j, x);
    }
  }
}

void
shade (struct image *dest, struct image *src)
{
  int  w = src->w;
  int  h = src->h;
  int  i, j;
  
  for (i=0; i<w; ++i) {
    for (j=0; j<h; ++j) {
      uchar  x = get_pixel (src, i, j);
      set_pixel (dest, i, j, reduce(x));
    }
  }

  g_free (weight);
  weight = weights (w, h);
  g_free (dest_mean);
  dest_mean = mean_values (dest);
  g_free (src_mean);
  src_mean = mean_values (src);
}

static uchar
select_best (double d0, double d64, double d128)
{
  if (d0 < d64 && d0 < d128)  return  0;
  if (d64 < d0 && d64 < d128)  return  64;
  return  128;
}

static void
diffuse_pix (struct image *dest, struct image *src,
	     int i, int j)
{
  double  d0, d64, d128;
  
  d0 = delta (dest, i, j, 0);
  d64 = delta (dest, i, j, 64);
  d128 = delta (dest, i, j, 128);
  set_pixel_and_mean (dest, i, j, select_best (d0, d64, d128));
}

static uchar
select_random (double d0, double d64, double d128, double beta)
{
  double  a, b, c, z, x;

  a = exp (-beta*d0);
  b = exp (-beta*d64);
  c = exp (-beta*d128);

  z = a+b+c;
  a /= z;  b /= z;  c /= z;
  if (! (a>0 && b>0 && c>0)) return  select_best (d0, d64, d128);

  x = jvrand_uniform (0, 1);
  if (x<a)  return  0;
  if (x<a+b)  return  64;
  return  128;
}

static void
gibbs_diffuse_pix (struct image *dest, struct image *src,
		   int i, int j, double beta)
{
  double  d0, d64, d128;
  
  d0 = delta (dest, i, j, 0);
  d64 = delta (dest, i, j, 64);
  d128 = delta (dest, i, j, 128);
  set_pixel_and_mean (dest, i, j, select_random(d0, d64, d128, beta));
}

void
diffuse (struct image *dest, struct image *src)
{
  int  w = src->w;
  int  h = src->h;
  int  i0, j0;

  for (i0=0; i0<2*RANGE; ++i0) {
    for (j0=0; j0<2*RANGE; ++j0) {
      int  i, j;
      
      for (i=i0; i<w; i+=2*RANGE) {
	for (j=j0; j<h; j+=2*RANGE) {
	  diffuse_pix (dest, src, i, j);
	}
      }
    }
  }
}

void
diffuse_rnd (struct image *dest, struct image *src)
{
  int  w = src->w;
  int  h = src->h;
  int  k;

  for (k=0; k<w*h; ++k) {
    int  i = jvrand_discrete (0, w);
    int  j = jvrand_discrete (0, h);
    diffuse_pix (dest, src, i, j);
  }
}

void
gibbs_diffuse (struct image *dest, struct image *src, double beta)
{
  int  w = src->w;
  int  h = src->h;
  int  i0, j0;

  for (i0=0; i0<2*RANGE; ++i0) {
    for (j0=0; j0<2*RANGE; ++j0) {
      int  i, j;
      
      for (i=i0; i<w; i+=2*RANGE) {
	for (j=j0; j<h; j+=2*RANGE) {
	  gibbs_diffuse_pix (dest, src, i, j, beta);
	}
      }
    }
  }
}

void
gibbs_diffuse_rnd (struct image *dest, struct image *src, double beta)
{
  int  w = src->w;
  int  h = src->h;
  int  k;

  for (k=0; k<w*h; ++k) {
    int  i = jvrand_discrete (0, w);
    int  j = jvrand_discrete (0, h);
    gibbs_diffuse_pix (dest, src, i, j, beta);
  }
}
